/*
 *  linux/fs/junkfs/super.c
 *
 * Copyright (C) September 13, 2008 and on
 * Maciej Żenczykowski (zenczykowski@gmail.com)
 *
 * Site Reliability Engineer
 * Google, Inc. (maze@google.com)
 *
 * PhD Student
 * Algorithmics Research Group
 * Theoretical Computer Science
 * Faculty of Mathematics and Computer Science
 * Jagiellonian University (maze@tcs.uj.edu.pl)
 *
 */

#include <linux/init.h>
#include <linux/module.h>
//#include <linux/stringify.h>

//#include <linux/slab.h>

//#include <linux/wait.h>

#include <linux/bio.h>

#include <linux/fs.h>

#include <linux/random.h>

#include "debug.h"
// printk_ln
// have it all buffer, than dump on printk_ln


static void rnd_uuid (void) {
	unsigned char uuid[16];
	int i;

	DBG_ENTER0();
	generate_random_uuid(uuid);
	printk(KERN_DEBUG "random uuid = [");
	for (i = 0; i < 15; ++i)
		printk("%02X", uuid[i]);
	printk("]\n");
	DBG_RETURN0();
}


static int test (int a, int b) {
	DBG_ENTER2(int, a, int, b);
	rnd_uuid();
	rnd_uuid();
	DBG_RETURN1(int, a + b);
}


/*
typedef struct file_system_type   struct_file_system_type;
typedef              const char   const_char;
typedef         struct vfsmount   struct_vfsmount;

typedef struct_file_system_type * struct_file_system_type_ptr;
typedef              const_char * const_char_ptr;
typedef                    void * void_ptr;
typedef         struct_vfsmount * struct_vfsmount_ptr;

static void printk_struct_file_system_type(struct_file_system_type p);
static void printk_const_char(const_char p);
static void printk_struct_vfsmount(struct_vfsmount p);

static void printk_struct_file_system_type_ptr(struct_file_system_type_ptr p);
static void printk_const_char_ptr(const_char_ptr p);
static void printk_void_ptr(void_ptr p);
static void printk_struct_vfsmount_ptr(struct_vfsmount_ptr p);

static int junkfs_get_sb(struct_file_system_type_ptr fs_type, int flags, const_char_ptr dev_name, void_ptr data, struct_vfsmount_ptr mnt)
{
	DBG_ENTER5(struct_file_system_type_ptr, fs_type, int, flags, const_char_ptr, dev_name, void_ptr, data, struct_vfsmount_ptr, mnt);
//      return get_sb_bdev(fs_type, flags, dev_name, data, ext2_fill_super, mnt);
	DBG_RETURN1(int, -ENOMEM);
}
*/


static long test_wq(void) {
	wait_queue_head_t * wq;
	int s;

	DBG_ENTER0();

	wq = kmalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
	if (IS_ERR(wq))
		return PTR_ERR(wq);

	DBG_MARK0();

	init_waitqueue_head(wq);

	DBG_MARK0();

	wake_up_all(wq);

	DBG_MARK0();

	printk(KERN_DEBUG "%016llx\n", get_jiffies_64());

//	s = wait_event_interruptible(wq, true);
	s = wait_event_interruptible_timeout(*wq, false, 1000);

	printk(KERN_DEBUG "%016llx\n", get_jiffies_64());

	DBG_MARK1(int, s);

	kfree(wq);

	DBG_RETURN1(long, 0L);
}


struct mz_t {
	wait_queue_head_t wq;
	int completed;
};


static void end_io_read(struct bio *bio, int err)
{
	struct mz_t * mzp;

	DBG_ENTER0();

	mzp = (struct mz_t *)bio->bi_private;
	mzp->completed = 1;
	bio_put(bio);
	wake_up(&mzp->wq);

	DBG_RETURN0();
}


#define SB_SIZE 512
static int junkfs_fill_super(struct super_block *sb, void *data, int silent)
{
	struct mz_t *mz;
	u8 *buf;
	struct bio *bio;
	int err;
	int s, y, x;

	DBG_ENTER0();

	mz = kmalloc(sizeof(struct mz_t), GFP_KERNEL);
	if (IS_ERR(mz)) {
		err = PTR_ERR(mz);
		goto out_mz_null;
	};

	init_waitqueue_head(&mz->wq);
	mz->completed = 0;

	buf = kmalloc(SB_SIZE, GFP_KERNEL);
	if (IS_ERR(buf)) {
		err = PTR_ERR(buf);
		goto out_buf_null;
	};

	bio = bio_alloc(GFP_KERNEL, 1);
	if (IS_ERR(bio)) {
		err = PTR_ERR(bio);
		goto out_bio_null;
	};

	bio->bi_bdev = sb->s_bdev;
	bio->bi_sector = 0; // first sector

	s = bio_add_page(bio, virt_to_page(buf), SB_SIZE, offset_in_page(buf));
	if (s == SB_SIZE) {
		bio->bi_end_io = end_io_read;
		bio->bi_private = mz;
		submit_bio(READ, bio);
		s = wait_event_interruptible(mz->wq, mz->completed);
		DBG_MARK1(int, s);
		for (y = 0; y < 32; ++y) {
			printk(KERN_INFO "%04X:", y * 16);
			for(x = 0; x < 16; ++x) {
				printk(" %02X", buf[y * 16 + x]);
			};
			printk("\n");
		};
	} else {
		DBG_MARK1(int, s);
		bio_put(bio);
	};
	err = -ENODATA;
out_bio_null:
	kfree(buf);
	buf = NULL;
out_buf_null:
	kfree(mz);
	mz = NULL;
out_mz_null:
	DBG_RETURN1(int, err);
}


static int junkfs_get_sb(struct file_system_type *fs_type, int flags, const char *dev_name, void *data, struct vfsmount *mnt)
{
	return get_sb_bdev(fs_type, flags, dev_name, data, junkfs_fill_super, mnt);
}

static struct file_system_type junk_fs_type = {
	.owner		= THIS_MODULE,
	.name		= "junkfs",
	.get_sb		= junkfs_get_sb,
	.kill_sb	= kill_block_super,
	.fs_flags	= FS_REQUIRES_DEV,
};


static int __init init_junk_fs(void) {
	int err;
	printk(KERN_INFO "- = - = - = - = - = -\n");
	DBG_ENTER0();

	test(5, 6);

	test_wq();

	err = register_filesystem(&junk_fs_type);
	DBG_MARK1(int, err);
	if (err)
		goto out;
	DBG_RETURN1(int, 0); // return 0;
out:
	DBG_RETURN1(int, err); // return err;
}


static void __exit exit_junk_fs(void) {
	DBG_ENTER0();
	unregister_filesystem(&junk_fs_type);
	DBG_RETURN0();
}


module_init(init_junk_fs)
module_exit(exit_junk_fs)

MODULE_AUTHOR("Maciej Żenczykowski");
MODULE_DESCRIPTION("The Junk Filesystem");
MODULE_LICENSE("GPL");
